/******************************************************************************************
    Copyright (C) 1997-2014 Andrew F. Neuwald, Cold Spring Harbor Laboratory
    and the University of Maryland School of Medicine.

    Permission is hereby granted, free of charge, to any person obtaining a copy of 
    this software and associated documentation files (the "Software"), to deal in the 
    Software without restriction, including without limitation the rights to use, copy, 
    modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so, subject to the 
    following conditions:

    The above copyright notice and this permission notice shall be included in all 
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR 
    PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE 
    LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT 
    OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
    OTHER DEALINGS IN THE SOFTWARE.

    For further information contact:
         Andrew F. Neuwald
         Institute for Genome Sciences and
         Department of Biochemistry & Molecular Biology
         University of Maryland School of Medicine
         801 West Baltimore St.
         BioPark II, Room 617
         Baltimore, MD 21201
         Tel: 410-706-6724; Fax: 410-706-1482; E-mail: aneuwald@som.umaryland.edu
 ******************************************************************************************/

/* pdis_class.c 08/17/99 
   Modified by Trevor Carlson (tcarlson@andrew.cmu.edu)
   The Original program can be found at:
     http://evolution.genetics.washington.edu/phylip.html

   Protein Distance: This program is used to compute the pairwise distances
  between proteins.  You supply the program with a char** of protiein sequences
  that are already aligned.  It returns a double array through the getDistace()
  accessor.  

  pdis_class(char**, Int4, Int4 ) - char** - char array of sequences.
                                    Int4   - number of sequences in this data set.
			      (2nd) Int4   - length of each sequence.
  double ** getDistance()         - returns the distance matrix for the data set.
                          **NOTE**- If you want to input this data into the nbr_class
			            you must increment the data first (d+1).

  NOTE: Progress output has been condensed into the printHeadings() function.  It is 
   still available and working, and has been carried over just in case.  It does 
   print to an output file is it is enabled.  Use printdata or printdis to control
   this and disance matrix output.
   Default condition: no output to files, only progress indicators to stdout.

  Bugs: No known.

*/
/* version 3.52c. (c) Copyright 1993 by Joseph Felsenstein and Mary Kuhner.
   Written by Mary Kuhner, Jon Yamato, Joseph Felsenstein, Akiko Fuseki,
   Sean Lamont, and Andrew Keeffe.
   Permission is granted to copy and use this program provided no fee is
   charged for it and provided that this copyright notice is not removed. */
#ifndef __pdis_class_h__
#define __pdis_class_h__

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "phylip.h"

#define epsilon         0.000001   /* a small number */
#define nmlngth         10

typedef enum {
  ala, arg, asn, asp, cys, gln, glu, gly, his, ileu, leu, lys, met, phe, pro,
  ser, thr, trp, tyr, val, del, stop, asx, glx, unk, quest
} aas;
typedef enum {
  universal, ciliate, mito, vertmito, flymito, yeastmito
} codetype;
typedef enum {
  chemical, hall, george
} cattype;

static double pameigs[] = {-0.022091252,-0.019297602, 0.000004760,-0.017477817,
                           -0.016575549,-0.015504543,-0.002112213,-0.002685727,
                           -0.002976402,-0.013440755,-0.012926992,-0.004293227,
                           -0.005356688,-0.011064786,-0.010480731,-0.008760449,
                           -0.007142318,-0.007381851,-0.007806557,-0.008127024
			   };

static double pamprobs[20][20] =
{
      {-0.01522976,-0.00746819,-0.13934468, 0.11755315,-0.00212101,
       0.01558456,-0.07408235,-0.00322387, 0.01375826, 0.00448826,
       0.00154174, 0.02013313,-0.00159183,-0.00069275,-0.00399898,
       0.08414055,-0.01188178,-0.00029870, 0.00220371, 0.00042546},
      {-0.07765582,-0.00712634,-0.03683209,-0.08065755,-0.00462872,
       -0.03791039,0.10642147,-0.00912185,0.01436308,-0.00133243,
        0.00166346,0.00624657,-0.00003363,-0.00128729,-0.00690319,
        0.17442028,-0.05373336,-0.00078751,-0.00038151,0.01382718},
      {-0.08810973,-0.04081786,-0.04066004,-0.04736004,-0.03275406,
      -0.03761164,-0.05047487, -0.09086213,-0.03269598,-0.03558015,
      -0.08407966,-0.07970977, -0.01504743,-0.04011920,-0.05182232,
      -0.07026991,-0.05846931, -0.01016998,-0.03047472,-0.06280511},
      { 0.02513756,-0.00578333, 0.09865453, 0.01322314,-0.00310665,
        0.05880899,-0.09252443,-0.02986539,-0.03127460, 0.01007539,
       -0.00360119,-0.01995024, 0.00094940,-0.00145868,-0.01388816,
        0.11358341,-0.12127513,-0.00054696,-0.00055627, 0.00417284},
      { 0.16517316,-0.00254742,-0.03318745,-0.01984173, 0.00031890,
       -0.02817810, 0.02661678,-0.01761215, 0.01665112, 0.10513343,
       -0.00545026, 0.01827470,-0.00207616,-0.00763758,-0.01322808,
       -0.02202576,-0.07434204, 0.00020593, 0.00119979,-0.10827873},
      { 0.16088826, 0.00056313,-0.02579303,-0.00319655, 0.00037228,
       -0.03193150, 0.01655305,-0.03028640, 0.01367746,-0.11248153,
        0.00778371, 0.02675579, 0.00243718, 0.00895470,-0.01729803,
       -0.02686964,-0.08262584, 0.00011794,-0.00225134, 0.09415650},
      {-0.01739295, 0.00572017,-0.00712592,-0.01100922,-0.00870113,
       -0.00663461,-0.01153857,-0.02248432,-0.00382264,-0.00358612,
       -0.00139345,-0.00971460,-0.00133312, 0.01927783,-0.01053838,
       -0.00911362,-0.01010908, 0.09417598, 0.01763850,-0.00955454},
      { 0.01728888, 0.01344211, 0.01200836, 0.01857259,-0.17088517,
        0.01457592, 0.01997839, 0.02844884, 0.00839403, 0.00196862,
        0.01391984, 0.03270465, 0.00347173,-0.01940984, 0.01233979,
        0.00542887, 0.01008836, 0.00126491,-0.02863042, 0.00449764},
      {-0.02881366,-0.02184155,-0.01566086,-0.02593764,-0.04050907,
       -0.01539603,-0.02576729,-0.05089606,-0.00597430, 0.02181643,
        0.09835597,-0.04040940, 0.00873512, 0.12139434,-0.02427882,
       -0.02945238,-0.01566867,-0.01606503, 0.09475319, 0.02238670},
      { 0.04080274,-0.02869626,-0.05191093,-0.08435843, 0.00021141,
        0.13043842, 0.00871530, 0.00496058,-0.02797641,-0.00636933,
        0.02243277, 0.03640362,-0.05735517, 0.00196918,-0.02218934,
       -0.00608972, 0.02872922, 0.00047619, 0.00151285, 0.00883489},
      {-0.02623824, 0.00331152, 0.03640692, 0.04260231,-0.00038223,
       -0.07480340,-0.01022492,-0.00426473, 0.01448116, 0.01456847,
        0.05786680, 0.03368691,-0.10126924,-0.00147454, 0.01275395,
        0.00017574,-0.01585206,-0.00015767,-0.00231848, 0.02310137},
      {-0.00846258,-0.01508106,-0.01967505,-0.02772004, 0.01248253,
       -0.01331243,-0.02569382,-0.04461524,-0.02207075, 0.04663443,
        0.19347923,-0.02745691, 0.02288515,-0.04883849,-0.01084597,
       -0.01947187,-0.00081675, 0.00516540,-0.07815919, 0.08035585},
      {-0.06553111, 0.09756831, 0.00524326,-0.00885098, 0.00756653,
        0.02783099,-0.00427042,-0.16680359, 0.03951331,-0.00490540,
        0.01719610, 0.15018204, 0.00882722,-0.00423197,-0.01919217,
       -0.02963619,-0.01831342,-0.00524338, 0.00011379,-0.02566864},
      {-0.07494341,-0.11348850, 0.00241343,-0.00803016, 0.00492438,
        0.00711909,-0.00829147, 0.05793337, 0.02734209, 0.02059759,
       -0.02770280, 0.14128338, 0.01532479, 0.00364307, 0.05968116,
       -0.06497960,-0.08113941, 0.00319445,-0.00104222, 0.03553497},
      { 0.05948223,-0.08959930, 0.03269977,-0.03272374,-0.00365667,
       -0.03423294,-0.06418925,-0.05902138, 0.05746317,-0.02580596,
        0.01259572, 0.05848832, 0.00672666, 0.00233355,-0.05145149,
        0.07348503, 0.11427955, 0.00142592,-0.01030651,-0.04862799},
      {-0.01606880, 0.05200845,-0.01212967,-0.06824429,-0.00234304,
        0.01094203,-0.07375538, 0.08808629, 0.12394822, 0.02231351,
       -0.03608265,-0.06978045,-0.00618360, 0.00274747,-0.01921876,
       -0.01541969,-0.02223856,-0.00107603,-0.01251777, 0.05412534},
      { 0.01688843, 0.05784728,-0.02256966,-0.07072251,-0.00422551,
       -0.06261233,-0.08502830, 0.08925346,-0.08529597, 0.01519343,
       -0.05008258, 0.10931873, 0.00521033, 0.02593305,-0.00717855,
        0.02291527, 0.02527388,-0.00266188,-0.00871160, 0.02708135},
      {-0.04233344, 0.00076379, 0.01571257, 0.04003092, 0.00901468,
        0.00670577, 0.03459487, 0.12420216,-0.00067366,-0.01515094,
        0.05306642, 0.04338407, 0.00511287, 0.01036639,-0.17867462,
       -0.02289440,-0.03213205, 0.00017924,-0.01187362,-0.03933874},
      { 0.01284817,-0.01685622, 0.00724363, 0.01687952,-0.00882070,
       -0.00555957, 0.01676246,-0.05560456,-0.00966893, 0.06197684,
       -0.09058758, 0.00880607, 0.00108629,-0.08308956,-0.08056832,
       -0.00413297, 0.02973107, 0.00092948, 0.07010111, 0.13007418},
      { 0.00700223,-0.01347574, 0.00691332, 0.03122905, 0.00310308,
        0.00946862, 0.03455040,-0.06712536,-0.00304506, 0.04267941,
       -0.10422292,-0.01127831,-0.00549798, 0.11680505,-0.03352701,
       -0.00084536, 0.01631369, 0.00095063,-0.09570217, 0.06480321}
    };

class pdis_class {

 public:

pdis_class();

pdis_class( unsigned char **,  
	             Int4   , 
	             Int4     );

~pdis_class();

double** getDistance();

 private:

FILE *infile, *outfile;
char *PDIS_OUTFILE;
Int4 spp, chars, datasets,ith;
double freqa, freqc, freqg, freqt, ttratio, xi, xv, ease, fracchange;
boolean weights, printdis, printdata, progress, mulsets, interleaved,
       ibmpc, vt52, ansi, basesequal, usepam, kimura, compdist;
codetype whichcode;
cattype whichcat;
Int4 *weight;
aas **node;
aas trans[4][4][4];
double pie[20];
Int4 cat[(Int4)val - (Int4)ala + 1], numaa[(Int4)val - (Int4)ala + 1];
double eig[20];
double prob[20][20], eigvecs[20][20];
double **d;

/* Local variables for makedists, propagated globally for c version: */
  double tt, p, dp, d2p, q, elambdat;

void distroyVars();
void openfile(FILE **fp, char *filename, const char *mode, char *perm );
void printHeadings();
void setupGlobals();
void inputnumbers();
void inputoptions();
aas  convdata( char );
void code();
void cats();
void maketrans();
void givens(double (*a)[20],
	    Int4 i, 
	    Int4 j, 
	    Int4 n,
	    double ctheta, 
	    double stheta,
	    boolean left );
void coeffs(double x, 
	    double y, 
	    double *c, 
	    double *s,
	    double accuracy);
void tridiag(double (*a)[20],
	     Int4 n,
	     double accuracy);
void shiftqr(double (*a)[20],
	     Int4 n,
	     double accuracy);
void qreigen(double (*prob_var)[20],
	     Int4 n);
void pameigen();
void predict(Int4 nb1,
	     Int4 nb2);
void makedists();
void startup();
};

#endif


















